/*
 * fg_ctf (attach module)
 *
 * [Flag]
 * Team0-X=x
 * Team0-Y=y
 * Team0-Radius=3
 * Team1-X=x
 * Team1-Y=y
 * Team1-Radius=3
 * Captures=3
 * RespawnTime=4000
 *
 * 2 flag spawns. your flag belongs to your team when it is in the spawn. when
 * you kill an enemy flag carrier the flag is nueted so you can tag it and
 * return it to your flag spawn. if you carry the enemy flag back to your flag
 * spawn and your flag is there you win a token. 3 tokens and you win the game.
 * if a enemy kills your flag carrier you can pickup the flag and continue back
 * to your base.
 *
 * 3/12/04 - started by smong
 *
 */

#include "asss.h"

struct rdata
{
	Arena *arena;
	int fid; /* fid/freq are same thing */
	int x, y; /* save messing with adata and spawns inside timer func */
};

struct spawn_t
{
	int x, y, r;
};

struct adata
{
	int on;
	int score[2];

	struct spawn_t data[2];
	int cfg_caps, cfg_respawntime;
};

local void ArenaAction(Arena *arena, int action);
local void PlayerAction(Player *p, int action, Arena *arena);
local void Kill(Arena *arena, Player *killer, Player *killed,
	int bounty, int flags, int *pts, int *green);
local void FlagPickup(Arena *arena, Player *p, int fid, int oldfreq,
	int carried);
local void FlagPos(Arena *arena, int fid, int x, int y, int freq);
local void Pppk(Player *p, byte *pkt, int len);

local void conf_extract_entries(ConfigHandle ch, struct adata *ad);
local int in_spawn_area(struct spawn_t *d, int x, int y);

local int respawn_timer(void *clos);
local void respawn_cleanup(void *clos);

local int adkey;

local Imodman     *mm;
local Iconfig     *cfg;
local Iflags      *flags;
local Ichat       *chat;
local Iarenaman   *aman;
local Imainloop   *ml;
local Inet        *net;

/* conf reader */

local void conf_extract_entries(ConfigHandle ch, struct adata *ad)
{
	ad->data[0].x = cfg->GetInt(ch, "Flag", "Team0-X", 497);
	ad->data[0].y = cfg->GetInt(ch, "Flag", "Team0-Y", 512);
	ad->data[0].r = cfg->GetInt(ch, "Flag", "Team0-Radius", 3);
	ad->data[1].x = cfg->GetInt(ch, "Flag", "Team1-X", 527);
	ad->data[1].y = cfg->GetInt(ch, "Flag", "Team1-Y", 512);
	ad->data[1].r = cfg->GetInt(ch, "Flag", "Team1-Radius", 3);
	ad->cfg_caps = cfg->GetInt(ch, "Flag", "Captures", 3);
	ad->cfg_respawntime = cfg->GetInt(ch, "Flag", "RespawnTime", 4000);

	if (ad->cfg_caps < 1)
		ad->cfg_caps = 1;

	if (ad->cfg_respawntime < 10)
		ad->cfg_respawntime = 10;
}

/* stuff */

local int in_spawn_area(struct spawn_t *d, int x, int y)
{
	x = x >> 4;
	y = y >> 4;
	x = x + d->r / 2;
	y = y + d->r / 2;
	if (x < d->x) return 0;
	if (y < d->y) return 0;
	if (x > d->x + d->r) return 0;
	if (y > d->y + d->r) return 0;
	return 1;
}

/* callbacks */

local void ArenaAction(Arena *arena, int action)
{
	struct adata *ad = P_ARENA_DATA(arena, adkey);

	if (action == AA_CREATE || action == AA_CONFCHANGED)
		conf_extract_entries(arena->cfg, ad);
}

local void PlayerAction(Player *p, int action, Arena *arena)
{
	if (action == PA_ENTERARENA)
	{
		struct adata *ad = P_ARENA_DATA(arena, adkey);
		chat->SendMessage(p, "ctf score: team0:%d team1:%d (%d captures"
			" needed)", ad->score[0], ad->score[1], ad->cfg_caps);
	}
}

local void Kill(Arena *arena, Player *killer, Player *killed,
	int bounty, int flags, int *pts, int *green)
{
	struct adata *ad = P_ARENA_DATA(arena, adkey);
	if (flags)
	{
		struct rdata *r = amalloc(sizeof(struct rdata));
		r->arena = arena;
		r->fid = killed->p_freq % 2;
		r->x = ad->data[r->fid].x;
		r->y = ad->data[r->fid].y;
		ml->SetTimer(respawn_timer, ad->cfg_respawntime, 1000, r,
			arena + r->fid);
	}
}

local void FlagPickup(Arena *arena, Player *p, int fid, int oldfreq,
	int carried)
{
	ml->CleanupTimer(respawn_timer, arena + fid, respawn_cleanup);

	if (fid == p->p_freq)
	{
		/* return to spawn */
		struct adata *ad = P_ARENA_DATA(arena, adkey);
		flags->MoveFlag(arena, fid, ad->data[fid].x, ad->data[fid].y,
			fid);
	}
}

local void FlagPos(Arena *arena, int fid, int x, int y, int freq)
{
	if (freq == -1 && fid <= 1)
	{
		/* move to spawn */
		struct adata *ad = P_ARENA_DATA(arena, adkey);
		flags->MoveFlag(arena, fid, ad->data[fid].x, ad->data[fid].y, fid);
	}
}

local void Pppk(Player *p, byte *pkt, int len)
{
	struct C2SPosition *pos = (struct C2SPosition *)pkt;
	Arena *arena = p->arena;
	struct adata *ad;

	if (!arena || len < 22)
		return;

	ad = P_ARENA_DATA(arena, adkey);
	if (ad->on)
	{
		ArenaFlagData *afd = flags->GetFlagData(arena);
		if (afd->flagcount >= 2)
		{
			int freq = p->p_freq % 2;
			int nmefreq = (p->p_freq + 1) % 2;

			/* first see if their flag is at home */
			if (in_spawn_area(&ad->data[freq],
				afd->flags[freq].x << 4,
				afd->flags[freq].y << 4) &&
				/* and if they are carrying the enemy flag */
				afd->flags[nmefreq].carrier == p &&
				/* now see if they are inside their flag spawn
				 * area */
				in_spawn_area(&ad->data[freq], pos->x, pos->y))
			{
				flags->ReleaseFlagData(arena);

				/* move enemy flag back to their spawn */
				flags->MoveFlag(arena, nmefreq,
					ad->data[nmefreq].x,
					ad->data[nmefreq].y, nmefreq);

				/* increment score */
				ad->score[freq]++;

				/* check win */
				if (ad->score[freq] >= ad->cfg_caps)
				{
					/* fixme: move to flag reset callback */
					chat->SendArenaSoundMessage(arena,
						SOUND_DING, "NOTICE: Game over."
						" Freq %d wins with %d "
						"captures.", freq,
						ad->score[freq]);
					ad->score[0] = 0;
					ad->score[1] = 0;
					/* fixme: points */
					flags->FlagVictory(arena, freq, 1);
				}
				else
				{
					chat->SendArenaSoundMessage(arena,
						SOUND_BEEP3, "Freq %d caps a "
						"flag! %d more caps to win.",
						freq, ad->cfg_caps - ad->score[freq]);
				}
			}
			else
			{
				flags->ReleaseFlagData(arena);
			}
		}
		else
		{
			flags->ReleaseFlagData(arena);
		}
	}
}

/* timers */

local int respawn_timer(void *clos)
{
	struct rdata *r = clos;

	flags->MoveFlag(r->arena, r->fid, r->x, r->y, r->fid);
	afree(r);

	return FALSE;
}

local void respawn_cleanup(void *clos)
{
	afree(clos);
}

EXPORT const char info_fg_ctf[] = "v1 smong <soinsg@hotmail.com>";

EXPORT int MM_fg_ctf(int action, Imodman *mm_, Arena *arena)
{
	if (action == MM_LOAD)
	{
		mm    = mm_;
		cfg   = mm->GetInterface(I_CONFIG,     ALLARENAS);
		flags = mm->GetInterface(I_FLAGS,       ALLARENAS);
		chat  = mm->GetInterface(I_CHAT,       ALLARENAS);
		aman  = mm->GetInterface(I_ARENAMAN,   ALLARENAS);
		ml    = mm->GetInterface(I_MAINLOOP,   ALLARENAS);
		net   = mm->GetInterface(I_NET,        ALLARENAS);
		if (!cfg || !flags || !chat || !aman || !ml || !net)
			return MM_FAIL;

		adkey = aman->AllocateArenaData(sizeof(struct adata));
		if (adkey == -1) return MM_FAIL;

		mm->RegCallback(CB_ARENAACTION, ArenaAction, ALLARENAS);
		net->AddPacket(C2S_POSITION, Pppk);

		return MM_OK;
	}
	else if (action == MM_UNLOAD)
	{
		ml->CleanupTimer(respawn_timer, NULL, respawn_cleanup);
		net->RemovePacket(C2S_POSITION, Pppk);
		mm->UnregCallback(CB_ARENAACTION, ArenaAction, ALLARENAS);

		aman->FreeArenaData(adkey);

		mm->ReleaseInterface(cfg);
		mm->ReleaseInterface(flags);
		mm->ReleaseInterface(chat);
		mm->ReleaseInterface(aman);
		mm->ReleaseInterface(ml);
		mm->ReleaseInterface(net);
		return MM_OK;
	}
	else if (action == MM_ATTACH)
	{
		struct adata *ad = P_ARENA_DATA(arena, adkey);
		ad->on = TRUE;

		/* support load-n-play on already created arenas */
		ArenaAction(arena, AA_CONFCHANGED);

		mm->RegCallback(CB_PLAYERACTION, PlayerAction, arena);
		mm->RegCallback(CB_KILL,         Kill,         arena);
		mm->RegCallback(CB_FLAGPICKUP,   FlagPickup,   arena);
		mm->RegCallback(CB_FLAGPOS,      FlagPos,      arena);
		return MM_OK;
	}
	else if (action == MM_DETACH)
	{
		struct adata *ad = P_ARENA_DATA(arena, adkey);
		ad->on = FALSE;

		mm->UnregCallback(CB_PLAYERACTION, PlayerAction, arena);
		mm->UnregCallback(CB_KILL,         Kill,         arena);
		mm->UnregCallback(CB_FLAGPICKUP,   FlagPickup,   arena);
		mm->UnregCallback(CB_FLAGPOS,      FlagPos,      arena);

		ml->CleanupTimer(respawn_timer, arena + 0, respawn_cleanup);
		ml->CleanupTimer(respawn_timer, arena + 1, respawn_cleanup);

		ArenaAction(arena, AA_DESTROY);
		return MM_OK;
	}

	return MM_FAIL;
}

